# frozen_string_literal: true

require 'rails'
require 'knock'
require 'active_support/core_ext/numeric/time'
require 'active_support/dependencies'
require 'orm_adapter'
require 'securerandom'

require 'negroni/version'

# Negroni extracts common authentication configuration to be used across the
# application.
module Negroni
  extend ActiveSupport::Autoload

  autoload :Configuration,  'negroni/configuration'
  autoload :Encryptor,      'negroni/encryptor'
  autoload :OmniAuth,       'negroni/omniauth'
  autoload :ParamFilter,    'negroni/param_filter'
  autoload :Resolver,       'negroni/resolver'
  autoload :TokenGenerator, 'negroni/token_generator'
  autoload :TokenNotFound,  'negroni/token_not_found'

  # Namespace for controllers
  module Controllers
    autoload :Helpers,            'negroni/controllers/helpers'
    autoload :TokenAuthenticable, 'negroni/controllers/token_authenticable'
  end

  # Namespace for mailer-related code
  module Mailers
    autoload :Helpers, 'negroni/mailers/helpers'
  end

  # @api private
  # To hold all modules
  ALL = [] # rubocop:disable Style/MutableConstant

  extend Configuration::Delegation

  class << self
    # @!group Private Configuration

    # Stores OmniAuth configurations
    # @return [Hash<Symbol,OmniAuth::Config>]
    attr_accessor :omniauth_configs

    # When true, enter in paranoid mode to avoid user enumeration.
    # @return [Boolean]
    attr_accessor :paranoid

    # Stores the token generator
    # @return [TokenGenerator]
    attr_accessor :token_generator

    # Stores the secret key
    # @return [String]
    attr_accessor :secret_key

    # @!attribute [rw] mailer
    #   @return [Mailer]

    # Returns the mailer
    # @return [Mailer]
    def mailer
      @mailer_ref.resolve
    end

    # Set the mailer class
    def mailer=(new_mailer)
      @mailer_ref = ref(new_mailer)
    end

    # @!endgroup

    # Yields the module for configuration. This is the standard way to configure
    # Negroni.
    #
    # @yield [Negroni]
    # @return [Void]
    def configure
      yield self
    end

    # List of OmniAuth provider that are registered
    # @return [Array<Symbol>]
    def omniauth_providers
      omniauth_configs.keys
    end

    # Register an OmniAuth provider.
    #
    # @param provider [Symbol] the OmniAuth provider to register
    # @param args [Object*] any additional arguments needed to configure and
    #   initialize the provider
    #
    # @example
    #   config.omniauth :github, APP_ID, APP_SECRET
    #
    # @return [Void]
    def omniauth(provider, *args)
      config = Negroni::OmniAuth::Config.new(provider, args)
      omniauth_configs[config.strategy_name.to_sym] = config
    end

    # Generate a friendly string randomly to be used as a token.
    #
    # @note: Taken from `Devise`.
    #
    # @param length [Integer] the length of the token.  Default: 20
    # @return [String] the generated token
    def friendly_token(length = 20)
      rlength = (length * 3) / 4
      SecureRandom.urlsafe_base64(rlength).tr('lIO0', 'sxyz')
    end

    # Securely compare two passwords
    #
    # @param a [String] the hashed password
    # @param b [String] the password to compare
    #
    # @return [Boolean] whether or not the passwords match
    def secure_compare(a, b)
      return false if a.blank? || b.blank? || a.bytesize != b.bytesize

      l = a.unpack "C#{a.bytesize}"

      res = 0
      b.each_byte { |byte| res |= byte ^ l.shift }
      res.zero?
    end

    private

    # Create a reference
    # @return [Resolver]
    def ref(arg)
      ActiveSupport::Dependencies.reference(arg)
      Resolver.new(arg)
    end
  end

  self.omniauth_configs = {}
  self.paranoid = false
  self.token_generator = nil
  self.mailer = 'Negroni::Mailer'

  class << self
    # @!group Module Registration

    # Register available negroni modules. For the standard modules that Negroni
    # provides, this method is called from lib/negroni/modules.rb. Third-party
    # modules need to be added explicitly using this method.
    #
    # @note that adding a module using this method does not cause it to be used
    #   in the authentication process. That requires that the module be listed in
    #   the arguments passed to the 'negroni' method in the model class
    #   definition.
    #
    # @param module_name [Symbol] the name of the module to register
    # @param options [Hash] a hash of options
    # @option options [String] :model the load path to a custom __model__ for
    #         this module (to autoload.)
    # @option options [Symbol, Boolean] :controller the name of an existing or
    #         custom __controller__ for this module.
    # @option options [Symbol, Boolean] :route the named __route__ helper for
    #         this module.
    #
    # All values which accept a boolean will have the same name as the given
    # module name.
    #
    # @return [Void]
    #
    # @example
    #   Negroni.register_module(:party_module)
    #   Negroni.register_module(:party_module, model: 'party_module/model')
    #   Negroni.register_module(:party_module, insert_at: 0)
    #
    def register_module(module_name, options = {})
      options.assert_valid_keys(:model, :controller, :insert_at)

      ALL.insert (options[:insert_at] || -1), module_name

      if (controller = options[:controller])
        register_controller(controller, module_name)
      end

      options[:model] && register_model(options[:model], module_name)
    end

    private

    def register_controller(controller, module_name)
      controller = (controller == true ? module_name : controller)
      CONTROLLERS[module_name] = controller
    end

    def register_model(model, module_name)
      path = (model == true ? "negroni/models/#{module_name}" : model)
      camelized = ActiveSupport::Inflector.camelize(module_name.to_s)
      Negroni::Models.send(:autoload, camelized.to_sym, path)
    end
  end
end

require 'negroni/models'
require 'negroni/modules'
require 'negroni/engine'
