# frozen_string_literal: true

module Negroni
  # `Configuration` encapsulates all of the configuration for Negroni,
  # keeping it out of the main module.
  class Configuration
    # The default email validation regex.
    DEFAULT_EMAIL_VALIDATION_REGEX = /\A[^@\s]+@[^@\s]+\z/i
    # /\A[\w+\-.]+@[a-z\d\-.]+\.[a-z]+\z/i

    # @!group Authenticable Configuration

    # Keys used when authenticating
    # @return [Array<Symbol>]
    attr_accessor :authentication_keys

    # Keys that should be treated as case-insensitive
    # @return [Array<Symbol>]
    attr_accessor :case_insensitive_keys

    # Keys that should have whitespace skipped
    # @return [Array<Symbol>]
    attr_accessor :strip_whitespace_keys

    # When true, send an email to notify password changes
    # @return [Boolean]
    attr_accessor :send_password_change_notification

    # How long before a token is expired. If nil is provided, token will
    # last forever.
    #
    # @return [ActiveSupport::Duration]
    attr_accessor :token_lifetime

    # The audience claim to identify the recipients that the token
    # is intended for.
    #
    # @return [Object]
    attr_accessor :token_audience

    # The name of the exception class that will be raised upon receiving an
    # invalid auth token.
    #
    # Default: {Negroni::TokenNotFound}
    #
    # @return [Class, String]
    attr_accessor :not_found_exception

    # @!group Validatable Configuration

    # Regular expression to validate emails
    # @return [RegExp]
    attr_accessor :email_regexp

    # Range validation for password
    # @return [Range]
    attr_accessor :password_length

    # @!group Encryption Configuration

    # The number of times to hash the password
    # @return [Integer]
    attr_accessor :stretches

    # Used to hash the password. Generate one with `rake secret`.
    # @return [String] the pepper used to hash the password
    attr_accessor :pepper

    # The algorithm used to encode the token.  Default: 'HS256'
    #
    # @return [String]
    attr_accessor :token_algorithm

    # The secret key that will be used for the token.  Default:
    # {Negroni::secret_key}.
    #
    # @return [String]
    attr_accessor :token_secret

    # An optional public key used to decode tokens.
    #
    # @return [String]
    attr_accessor :token_public_key

    # @!group Lockable Configuration

    # Defines which strategy will be used to lock an account.
    #
    #   * `:failed_attempts` = Locks an account after a number of failed
    #           attempts.
    #   * `:none`            = No lock strategy. You should handle locking by
    #           yourself.
    #
    # @return [Symbol] the name of the lock strategy.
    attr_accessor :lock_strategy

    # Defines which key will be used when locking and unlocking an account
    # @return [Array<Symbol>]
    attr_accessor :unlock_keys

    # Defines which strategy can be used to unlock an account.
    # Valid values: `:email,` `:time,` `:both`
    # @return [Symbol]
    attr_accessor :unlock_strategy

    # Number of authentication tries before locking an account
    # @return [Integer]
    attr_accessor :maximum_attempts

    # Time interval to unlock the account if `:time` is defined as
    # `unlock_strategy`.
    #
    # @return [ActiveSupport::Duration]
    attr_accessor :unlock_in

    # Defines which key will be used when recovering the password for an account
    # @return [Array<Symbol>]
    attr_accessor :reset_password_keys

    # Time interval you can reset your password with a reset password key
    # @return [ActiveSupport::Duration]
    attr_accessor :reset_password_within

    # @!group Mailer Configuration

    # The sender for all mailers
    # @return [String] the email address for the sender
    attr_accessor :mailer_sender

    # The class `Mailer` should inherit from (`'ActionMailer::Base'` by default)
    # @return [String]
    attr_accessor :parent_mailer

    # @!group Controller Configuration

    # The class `Negroni::BaseController` should inherit from.
    #
    # Default: ActionController::API
    #
    # @return [String, Class]
    attr_accessor :parent_controller

    # @!endgroup

    # Create a new instance of `Configuration`, using default values for all
    # attributes.
    def initialize # rubocop:disable Metrics/AbcSize,MethodLength
      @authentication_keys = [:email]
      @case_insensitive_keys = [:email]
      @strip_whitespace_keys = [:email]
      @send_password_change_notification = false
      @token_lifetime = 1.day
      @token_algorithm = 'HS256'
      @not_found_exception = 'Negroni::TokenNotFound'
      @email_regexp = DEFAULT_EMAIL_VALIDATION_REGEX
      @password_length = 8..72
      @stretches = 11
      @pepper = nil
      @lock_strategy = :failed_attempts
      @unlock_keys = [:email]
      @unlock_strategy = :both
      @maximum_attempts = 20
      @unlock_in = 1.hour
      @reset_password_keys = [:email]
      @reset_password_within = 6.hours
      @mailer_sender = nil
      @parent_mailer = 'ActionMailer::Base'
      @parent_controller = 'ActionController::API'
    end

    # `Delegation` adds methods to the including or extending class to delegate
    # parameters to an instance of {Configuration}.
    #
    # Additionally, it provides a method `{#configuration}`, which is simply a
    # lazily-instantiated instance of {Configuration}.
    #
    module Delegation
      # Create delegation method for configuration
      #
      def self.config_delegator(*attrs)
        attrs.each do |attribute|
          class_eval <<-METHOD, __FILE__, __LINE__ + 1
            def #{attribute}
              configuration.#{attribute}
            end

            def #{attribute}=(new_value)
              configuration.#{attribute} = new_value
            end
          METHOD
        end
      end

      # The configuration object
      # @return [Configuration]
      def configuration
        @configuration ||= Configuration.new
      end

      # @!macro [attach] cfg.delegator
      #   Delegates `$1` to #configuration.
      #
      #   {render:Configuration#$1}
      #
      config_delegator :authentication_keys
      config_delegator :case_insensitive_keys
      config_delegator :strip_whitespace_keys
      config_delegator :send_password_change_notification
      config_delegator :token_lifetime
      config_delegator :token_audience
      config_delegator :not_found_exception
      config_delegator :email_regexp
      config_delegator :password_length
      config_delegator :stretches
      config_delegator :pepper
      config_delegator :token_algorithm
      config_delegator :token_secret
      config_delegator :token_public_key
      config_delegator :lock_strategy
      config_delegator :unlock_keys
      config_delegator :unlock_strategy
      config_delegator :maximum_attempts
      config_delegator :unlock_in
      config_delegator :reset_password_keys
      config_delegator :reset_password_within
      config_delegator :mailer_sender
      config_delegator :parent_mailer
      config_delegator :parent_controller
    end
  end
end
