# frozen_string_literal: true

module Negroni
  module Models
    # The `Base` module contains methods that should be included in all Negroni
    # models.  It holds common settings for all authentication, as well as some
    # shared behavior across all modules.
    #
    # ## Options
    #
    # Base adds the following options:
    #
    #   * `authentication_keys`: parameters used for authentication.
    #      Default [:email].
    #
    #   * `strip_whitespace_keys`: keys from which whitespace should be
    #      stripped prior to validation.  Default [:email].
    #
    module Base
      extend ActiveSupport::Concern

      # Blacklisted attributes for serialization.
      SERIALIZATION_BLACKLIST = [
        :password_digest, :reset_password_token, :reset_password_sent_at,
        :password_salt, :confirmation_token, :confirmed_at,
        :confirmation_sent_at, :failed_attempts, :unlock_token, :locked_at
      ].freeze

      included do
        class_attribute :negroni_modules, instance_writer: false
        self.negroni_modules ||= []

        before_validation :downcase_keys
        before_validation :strip_whitespace
      end

      # Required fields for this module (_none_).
      def self.required_fields(_klass)
        []
      end

      # Returns whether or not the including class is active for
      # authentication.  This method is primarily intended to be overriden by
      # other modules, including `Lockable`.
      #
      # Yields and returns the return value of the given block, if a block is
      # given.  Otherwise returns true.
      #
      # @return [Boolean]
      def valid_for_auth?
        block_given? ? yield : true
      end

      # The message that will be populated if an invalid record attempts to
      # authenticate.
      #
      # @return [Symbol] the key of the translation to look up
      def unauthenticated_message
        :invalid
      end

      # Returns whether or not the including class is active for
      # authentication.  This method is primarily intended to be overriden by
      # other modules, including `Lockable`.
      #
      # @return [Boolean]
      def active_for_auth?
        true
      end

      # The message that will be populated if an inactive record attempts to
      # authenticate.
      #
      # @return [Symbol] the key of the translation to look up
      def inactive_message
        :inactive
      end

      # Redefine serializable_hash in models for more secure defaults.  By
      # default, it removes from the serializable model all attributes that are
      # __not__ accessible. You can remove this default by using :force_except
      # and passing a new list of attributes you want to exempt. All attributes
      # given to :except will simply add names to exempt to Negroni internal
      # list.
      def serializable_hash(options = nil)
        options ||= {}
        options[:except] = Array(options[:except])

        if options[:force_except]
          options[:except].concat Array(options[:force_except])
        else
          options[:except].concat SERIALIZATION_BLACKLIST
        end

        super(options)
      end

      # Redefine inspect using serializable_hash, to ensure we don't accidentally
      # leak passwords into exceptions.
      def inspect
        inspection = serializable_hash.collect do |k, v|
          value = if respond_to?(:attribute_for_inspect)
                    attribute_for_inspect(k)
                  else
                    v.inspect
                  end

          "#{k}: #{value}"
        end

        "#<#{self.class} #{inspection.join(', ')}>"
      end

      protected

      def auth_mailer
        Negroni.mailer
      end

      # This is an internal method called every time Negroni needs to send a
      # notification/mail. This can be overridden if you need to customize the
      # e-mail delivery logic. For instance, if you are using a queue to deliver
      # e-mails (delayed job, sidekiq, resque, etc), you must add the delivery
      # to the queue just after the transaction was committed. To achieve this,
      # you can override send_auth_notification to store the deliveries
      # until the after_commit callback is triggered:
      #
      # @example
      #     class User
      #       negroni :authenticable, :confirmable
      #
      #       after_commit :send_pending_notifications
      #
      #       protected
      #
      #       def send_auth_notification(notification, *args)
      #         # If the record is new or changed then delay the
      #         # delivery until the after_commit callback otherwise
      #         # send now because after_commit will not be called.
      #         if new_record? || changed?
      #           pending_notifications << [notification, args]
      #         else
      #           message = auth_mailer.send(notification, self, *args)
      #           Remove once we move to Rails 4.2+ only.
      #           if message.respond_to?(:deliver_now)
      #             message.deliver_now
      #           else
      #             message.deliver
      #           end
      #         end
      #       end
      #
      #       def send_pending_notifications
      #         pending_notifications.each do |notification, args|
      #           message = auth_mailer.send(notification, self, *args)
      #           Remove once we move to Rails 4.2+ only.
      #           if message.respond_to?(:deliver_now)
      #             message.deliver_now
      #           else
      #             message.deliver
      #           end
      #         end
      #
      #         # Empty the pending notifications array because the
      #         # after_commit hook can be called multiple times which
      #         # could cause multiple emails to be sent.
      #         pending_notifications.clear
      #       end
      #
      #       def pending_notifications
      #         \@pending_notifications ||= []
      #       end
      #     end
      #
      def send_auth_notification(notification, *args)
        message = auth_mailer.send(notification, self, *args)
        message.deliver_now
      end

      # @!group Callbacks

      # Downcase keys before validation
      def downcase_keys
        self.class.case_insensitive_keys.each { |k| _apply(k, :downcase) }
      end

      # Strip whitespace from requested keys before validation
      def strip_whitespace
        self.class.strip_whitespace_keys.each { |k| _apply(k, :strip) }
      end

      # @!endgroup Callbacks

      private

      # Apply `method` to `attr`
      def _apply(attr, method)
        if self[attr]
          self[attr] = self[attr].try(method)
        elsif respond_to?(attr) && respond_to?("#{attr}=")
          new_value = send(attr).try(method)
          send("#{attr}=", new_value)
        end
      end

      # Class Methods for the `Base` module.
      module ClassMethods
        Negroni::Models.config(self,
                               :authentication_keys,
                               :strip_whitespace_keys,
                               :case_insensitive_keys)

        # Find first record based on conditions given (ie by the sign in form).
        # This method is always called during an authentication process but
        # it may be wrapped as well. For instance, database authenticable
        # provides a `find_for_database_authentication` that wraps a call to
        # this method. This allows you to customize both database authenticable
        # or the whole authenticate stack by customize `find_for_authentication.`
        #
        # Overwrite to add customized conditions, create a join, or maybe use a
        # namedscope to filter records while authenticating.
        #
        # @example
        #   def self.find_for_authentication(tainted_conditions)
        #     find_first_by_auth_conditions(tainted_conditions, active: true)
        #   end
        #
        # Finally, notice that Negroni also queries for users in other scenarios
        # besides authentication, for example when retrieving an user to send
        # an e-mail for password reset. In such cases, find_for_authentication
        # is not called.
        def find_for_authentication(tainted_conditions)
          find_first_by_auth_conditions(tainted_conditions)
        end

        # Find the first record, given a set of `tainted_conditions`, and
        # options.
        #
        # @param tainted_conditions [Hash, ActionDispatch::Parameters]
        #        the conditions used to find the record
        # @param options [Hash] additional conditions and options for the find
        #        operation
        #
        # @return [Object] the first record returned from the database
        def find_first_by_auth_conditions(tainted_conditions, options = {})
          to_adapter.find_first(
            _param_filter.filter(tainted_conditions).merge(options)
          )
        end

        # Find or initialize a record setting an error if it can't be found.
        def find_or_initialize_with_error_by(attr, value, error = :invalid)
          find_or_initialize_with_errors([attr], { attr => value }, error)
        end

        # Find or initialize a record with a group of attributes, based on a
        # list of required attributes
        def find_or_initialize_with_errors(required, attrs, error = :invalid)
          attrs = _indifferently(required, attrs).delete_if { |_, v| v.blank? }

          if attrs.size == required.size
            record = find_first_by_auth_conditions(attrs)
          end

          unless record
            record = new

            required.each do |key|
              value = attrs[key]
              record.send("#{key}=", value)
              record.errors.add(key, value.present? ? error : :blank)
            end
          end

          record
        end

        # Finds an entity of the including class by a token auth request.
        # This allows users to sign in with either an email address or
        # thier username.
        #
        # @param request [Request] The request which contains the sign in params.
        #
        # @return [Object, nil] The found entity, or `nil` if one was not found.
        def from_token_request(request)
          # Bail if there is no `auth` param
          return nil unless (auth_params = request.params['auth'])

          # find_first_by_auth_conditions(auth_params)

          authentication_keys.each do |key|
            if (found_key = auth_params[key.to_s])
              return to_adapter.find_first(key => found_key)
            end
          end
        end

        protected

        def _param_filter
          @_param_filter ||= Negroni::ParamFilter.new(
            case_insensitive_keys, strip_whitespace_keys
          )
        end

        private

        def _indifferently(required, attributes)
          if attributes.respond_to?(:permit!)
            attributes.slice(*required).permit!.to_h.with_indifferent_access
          else
            attributes.with_indifferent_access.slice(*required)
          end
        end
      end
    end
  end
end
