# frozen_string_literal: true

module Negroni
  module Models
    # Handles blocking a user access after a certain number of attempts.
    # Lockable accepts two different strategies to unlock a user after it's
    # blocked: email and time. The former will send an email to the user when
    # the lock happens, containing a link to unlock its account. The second will
    # unlock the user automatically after some configured time (ie 2.hours).
    # It's also possible to set up lockable to use both email and time
    # strategies.
    #
    # ## Options
    #
    # Lockable adds the following options to `negroni`:
    #
    #   * `maximum_attempts`: how many attempts should be accepted before
    #      blocking the user.
    #   * `lock_strategy`: lock the user account by :failed_attempts or :none.
    #   * `unlock_strategy`: unlock the user account by :time, :email, :both or
    #      :none.
    #   * `unlock_in`: the time you want to lock the user after to lock happens.
    #                  Only available when unlock_strategy is :time or :both.
    #   * `unlock_keys`: the keys you want to use when locking and unlocking.
    #
    module Lockable
      extend ActiveSupport::Concern

      # @!method lock_strategy_enabled?(strategy)
      #   @return [Boolean] if the passed strategy is enabled
      delegate :lock_strategy_enabled?,
               :unlock_strategy_enabled?,
               to: 'self.class'

      # Required fields
      def self.required_fields(klass)
        att = []
        att << :failed_attempts if klass.lock_strategy_enabled? :failed_attempts
        att << :locked_at if klass.unlock_strategy_enabled? :time
        att << :unlock_token if klass.unlock_strategy_enabled? :email
        att
      end

      # Lock a user setting its locked_at to actual time.
      #
      # @param opts [Hash] hash of options
      # @option opts [Boolean] :send_instructions pass `false` to not send the
      #         email
      def lock_access!(opts = {})
        self.locked_at = Time.now.utc

        if unlock_strategy_enabled?(:email) &&
           opts.fetch(:send_instructions, true)
          send_unlock_instructions
        else
          save(validate: false)
        end
      end

      # Unlock a user, clearing `locked_at` and `failed_attempts`.
      def unlock_access!
        self.locked_at = nil
        self.failed_attempts = 0 if respond_to?(:failed_attempts=)
        self.unlock_token = nil if respond_to?(:unlock_token=)
        save(validate: false)
      end

      # Returns true if the user has access locked
      # @return [Boolean]
      def access_locked?
        locked_at && !lock_expired?
      end

      # send instructions to unlock
      def send_unlock_instructions
        raw, enc = Negroni.token_generator.generate(self.class, :unlock_token)
        self.unlock_token = enc

        save(validate: false) # rubocop:disable Rails/SaveBang
        send_auth_notification(:unlock_instructions, raw, {})
        raw
      end

      # Resend instructions if the user is locked
      def resend_unlock_instructions
        if_access_locked { send_unlock_instructions }
      end

      # Override active_for_auth? for locking purposes
      def active_for_auth?
        super && !access_locked?
      end

      # Override for locking purposes
      def inactive_message
        access_locked? ? :locked : super
      end

      # rubocop:disable Metrics/CyclomaticComplexity,PerceivedComplexity

      # Override for locking purposes
      def valid_for_auth?
        return super unless persisted? &&
                            lock_strategy_enabled?(:failed_attempts)

        unlock_access! if lock_expired?

        return true if super && !access_locked?

        self.failed_attempts ||= 0
        self.failed_attempts += 1

        if attempts_exceeded?
          lock_access! unless access_locked?
        else
          save(validate: false)
        end

        false
      end

      # Override for locking
      def unauthenticated_message
        if Negroni.paranoid
          super
        elsif access_locked? ||
              (lock_strategy_enabled?(:failed_attempts) && attempts_exceeded?)
          :locked
        elsif lock_strategy_enabled?(:failed_attempts) &&
              last_attempt? && self.class.last_attempt_warning
          :last_attempt
        else
          super
        end
      end

      protected

      def attempts_exceeded?
        failed_attempts >= self.class.maximum_attempts
      end

      def last_attempt?
        failed_attempts == self.class.maximum_attempts - 1
      end

      def lock_expired?
        return false unless unlock_strategy_enabled?(:time)
        locked_at && locked_at < self.class.unlock_in.ago
      end

      def if_access_locked
        if access_locked?
          yield
        else
          errors.add(Negroni.unlock_keys.first, :not_locked)
          false
        end
      end

      # Class Methods for Lockable
      module ClassMethods
        # @private
        BOTH_STRATEGIES = [:time, :email].freeze
        private_constant :BOTH_STRATEGIES

        Negroni::Models.config(self, :maximum_attempts, :lock_strategy,
                               :unlock_strategy, :unlock_in, :unlock_keys,
                               :last_attempt_warning)

        # Attempt to find a user by its unlock keys. If a record is found, send
        # new unlock instructions to it. If not user is found, returns a new
        # user with an email not found error.  Options must contain the user's
        # unlock keys Checks if a given lock strategy is enabled.
        def send_unlock_instructions(attributes = {})
          lockable = find_or_initialize_with_errors(
            unlock_keys, attributes, :not_found
          )
          lockable.resend_unlock_instructions if lockable.persisted?
          lockable
        end

        # Find a user by its unlock token and try to unlock it.
        # If no user is found, returns a new user with an error.
        # If the user is not locked, creates an error for the user
        # Options must have the unlock_token
        def unlock_access_by_token(token)
          orig = token
          token = Negroni.token_generator.digest(self, :unlock_token, token)

          lockable = find_or_initialize_with_error_by(:unlock_token, token)
          lockable.unlock_access! if lockable.persisted?
          lockable.unlock_token = orig
          lockable
        end

        # @param strategy [Symbol] the name of the strategy.
        #
        # @return [Boolean] if the strategy is enabled.
        def lock_strategy_enabled?(strategy)
          lock_strategy == strategy
        end

        # Checks if a given unlock strategy is enabled.
        #
        # @param strategy [Symbol] the name of the strategy.
        #
        # @return [Boolean] if the strategy is enabled.
        def unlock_strategy_enabled?(strategy)
          unlock_strategy == strategy ||
            (unlock_strategy == :both && BOTH_STRATEGIES.include?(strategy))
        end
      end
    end
  end
end
