# frozen_string_literal: true

module Negroni
  module Models
    # Recoverable takes care of resetting the user password and send reset
    # instructions.
    #
    # ## Required attributes
    #
    # Recoverable requires that the including class has the following attributes:
    #
    #   * `reset_password_sent_at`: [DateTime]
    #   * `reset_password_token`: [String]
    #
    # ## Options
    #
    # Recoverable adds the following class options:
    #
    #   * `reset_password_keys`: the keys you want to use when recovering the
    #                            password for an account
    #   * `reset_password_within`: the time period within which the password must
    #                              be reset or the token expires.
    #   * `sign_in_after_reset_password`: whether or not to sign in the user
    #                                     automatically after a password reset.
    #
    # @example
    #   # resets the user password and save the record, true if valid passwords
    #   # are given, otherwise false
    #   User.find(1).reset_password('password123', 'password123')
    #
    # @example
    #   # creates a new token and send it with instructions about how to reset
    #   # the password
    #   User.find(1).send_reset_password_instructions
    #
    module Recoverable
      extend ActiveSupport::Concern

      # Required fields for Recoverable
      def self.required_fields(_klass = nil)
        [:reset_password_sent_at, :reset_password_token]
      end

      included do
        before_update :clear_reset_password_token,
                      if: :clear_reset_password_token?
      end

      # Update password saving the record and clearing token. Returns true if
      # the passwords are valid and the record was saved, false otherwise.
      #
      # @param new_password [String] the new password
      # @param new_password_confirmation [String] confirmation of the password
      #
      # @return [Boolean] if the operation was successful
      def reset_password(new_password, new_password_confirmation)
        self.password = new_password
        self.password_confirmation = new_password_confirmation

        save
      end

      # Resets reset password token and send reset password instructions by
      # email.  Returns the token sent in the e-mail.
      #
      # @return [String] the token that was sent in the email.
      def send_reset_password_instructions
        token = set_reset_password_token
        send_reset_password_instructions_notification(token)

        token
      end

      # Checks if the reset password token sent is within the limit time.  We do
      # this by calculating if the difference between today and the sending date
      # does not exceed the confirm in time configured.  Returns true if the
      # resource is not responding to reset_password_sent_at at all.
      #
      # @return [Boolean] if the period is valid.
      #
      # reset_password_within is a model configuration, must always be an integer
      # value.
      #
      # @example
      #   # reset_password_within = 1.day and reset_password_sent_at = today
      #   reset_password_period_valid?   # => true
      #
      # @example
      #   # reset_password_within = 5.days; reset_password_sent_at = 4.days.ago
      #   reset_password_period_valid?   # => true
      #
      # @example
      #   # reset_password_within = 5.days; reset_password_sent_at = 5.days.ago
      #   reset_password_period_valid?   # => false
      #
      # @example
      #   # reset_password_within = 0.days
      #   reset_password_period_valid?   # will always return false
      #
      def reset_password_period_valid?
        reset_password_sent_at &&
          reset_password_sent_at.utc >= self.class.reset_password_within.ago.utc
      end

      protected

      # Removes `reset_password_token`
      def clear_reset_password_token
        self.reset_password_token = nil
        self.reset_password_sent_at = nil
      end

      # Sets the `reset_password_token` for the record
      def set_reset_password_token
        raw, encoded = Negroni.token_generator.generate(self.class,
                                                        :reset_password_token)

        self.reset_password_token = encoded
        self.reset_password_sent_at = Time.now.utc

        save(validate: false) # rubocop:disable Rails/SaveBang
        raw
      end

      def send_reset_password_instructions_notification(token)
        send_auth_notification(:reset_password_instructions, token, {})
      end

      def clear_reset_password_token?
        auth_keys_changed = self.class.authentication_keys.any? do |attribute|
          respond_to?("#{attribute}_changed?") && send("#{attribute}_changed?")
        end

        !new_record? && (auth_keys_changed || password_digest_changed?)
      end

      # Class Methods for `Recoverable`
      module ClassMethods
        # Attempt to find a user by password reset token. If a user is found,
        # return it.  If not, return `nil`.
        def with_reset_password_token(token)
          reset_password_token = Negroni.token_generator.digest(
            self, :reset_password_token, token
          )
          to_adapter.find_first(reset_password_token: reset_password_token)
        end

        # Attempt to find a user by its email. If a record is found, send new
        # password instructions to it. If user is not found, returns a new user
        # with an email not found error.
        #
        # Attributes must contain the user's email
        def send_reset_password_instructions(attributes = {})
          recoverable = find_or_initialize_with_errors(reset_password_keys,
                                                       attributes,
                                                       :not_found)

          recoverable.send_reset_password_instructions if recoverable.persisted?
          recoverable
        end

        # rubocop:disable Metrics/MethodLength

        # Attempt to find a user by its reset_password_token to reset its
        # password. If a user is found and token is still valid, reset its
        # password and automatically try saving the record. If not user is
        # found, returns a new user containing an error in reset_password_token
        # attribute.
        #
        # Attributes must contain reset_password_token, password and confirmation
        def reset_password_by_token(attributes = {})
          original = attributes[:reset_password_token]
          token = Negroni.token_generator.digest(
            self, :reset_password_token, original
          )

          recoverable = find_or_initialize_with_error_by(:reset_password_token,
                                                         token)

          if recoverable.persisted?
            if recoverable.reset_password_period_valid?
              recoverable.reset_password(attributes[:password],
                                         attributes[:password_confirmation])
            else
              recoverable.errors.add(:reset_password_token, :expired)
            end
          end

          if recoverable.reset_password_token.present?
            recoverable.reset_password_token = original
          end

          recoverable
        end

        Negroni::Models.config(self,
                               :reset_password_keys,
                               :reset_password_within)
      end
    end
  end
end

# rubocop:enable all
