# frozen_string_literal: true

module Negroni
  module Models
    # Validatable creates all needed validations for a user email and password.
    # It's optional, given you may want to create the validations by yourself.
    # Automatically validate if the email is present, unique and its format is
    # valid. Also tests presence of password, confirmation and length.
    #
    # ## Options
    #
    # Validatable adds the following options to devise_for:
    #
    #   * `email_regexp`: the regular expression used to validate e-mails
    #   * `password_length`: a range expressing password length. Default: 8..72
    #
    module Validatable
      extend ActiveSupport::Concern

      # Required fields for :validatable
      def self.required_fields(_klass = nil)
        []
      end

      included do
        raise RuntimeError unless respond_to?(:validates)

        validates :email, presence: true, if: :email_required?

        with_options allow_blank: true, if: :email_changed? do |val|
          val.validates :email, uniqueness: true
          val.validates :email, format: { with: email_regexp }
        end

        validates :password, presence: true, if: :password_required?

        validates :password, confirmation: true, if: :password_required?

        validates :password, length: { in: password_length }, allow_blank: true

        validates :password_confirmation, presence: true,
                                          on: [:create, :update],
                                          if: :password_required?
      end

      protected

      def password_required?
        !persisted? || !password.nil? || !password_confirmation.nil?
      end

      # Override this method if an email is not required for your model
      def email_required?
        true
      end

      # Class methods for Validations
      module ClassMethods
        Negroni::Models.config(self, :email_regexp, :password_length)
      end
    end
  end
end
