# frozen_string_literal: true

module Negroni
  module OmniAuth
    # Raised if a strategy is unknown or not found.
    class UnknownStrategy < NameError
      # Raise a new error for `strategy`.
      def initialize(strategy)
        @strategy = strategy
        super("Could not find a strategy named `#{strategy}`.  " \
              "Please ensure it is `require`d or explicitly set using " \
              "the :strategy_class option.")
      end
    end

    # Config encapsulates OmniAuth configuration for Negroni
    class Config
      # @!attribute [rw] strategy
      #   @return [Symbol] the strategy for OmniAuth
      attr_accessor :strategy

      # @!attribute [r] args
      #   @return [Array<Object>] arguments needed to configure the provider
      attr_reader :args

      # @!attribute [r] options
      #   @return [Hash] options needed to configure the provider
      attr_reader :options

      # @!attribute [r] provider
      #   @return [Symbol] the name of the OmniAuth provider
      attr_reader :provider

      # @!attribute [r] strategy_name
      #   @return [String, Symbol] the name of the strategy, if different from
      #     `provider`
      attr_reader :strategy_name

      # Create a new OmniAuth::Config
      #
      # @param provider [Symbol] the name of the provider to use
      # @param args [Array<Object>] any arguments to pass to the provider
      #
      def initialize(provider, args)
        @provider       = provider
        @args           = args
        @options        = @args.last.is_a?(Hash) ? @args.last : {}
        @strategy       = nil
        @strategy_name  = options[:name] || @provider
        @strategy_class = options.delete(:strategy_class)
      end

      # @!attribute [r] strategy_class
      #   @return [Class] the class for the OmniAuth strategy

      # @return [Class] the class for the OmniAuth strategy
      def strategy_class
        @strategy_class ||= find_strategy || autoload_strategy
      end

      # Find a the strategy, using the name provided
      def find_strategy
        ::OmniAuth.strategies.find do |klass|
          klass.to_s =~ /#{::OmniAuth::Utils.camelize(strategy_name)}$/ ||
            klass.default_options[:name] == strategy_name
        end
      end

      # Autoload a strategy
      def autoload_strategy
        name = ::OmniAuth::Utils.camelize(provider.to_s)

        if ::OmniAuth::Strategies.const_defined?(name)
          return ::OmniAuth::Strategies.const_get(name)
        end

        raise UnknownStrategy, name
      end
    end
  end
end
