# frozen_string_literal: true

module Negroni
  # ParamFilter is responsible for filtering blacklisted parameters from a
  # request for validation and authentication.
  class ParamFilter
    # Creates a new instance of `ParamFilter`.
    #
    # @param case_insensitive_keys [Array<Symbol>] keys which are not case
    #   sensitive
    # @param strip_whitespace_keys [Array<Symbol>] keys which should have
    #   whitespace stripped
    #
    def initialize(case_insensitive_keys, strip_whitespace_keys)
      @case_insensitive_keys = case_insensitive_keys || []
      @strip_whitespace_keys = strip_whitespace_keys || []
    end

    # Filter the `conditions` based on \@case_insensitive_keys and
    #   \@strip_whitespace_keys.
    #
    # @param conditions [Hash] the conditions hash to filter
    #
    # @return [Hash] the filtered `conditions` hash
    def filter(conditions)
      conditions = stringify_params(conditions.dup.to_h)

      conditions.merge! filtered_hash_by_meth_for_keys(conditions.dup,
                                                       :downcase,
                                                       @case_insensitive_keys)

      conditions.merge! filtered_hash_by_meth_for_keys(conditions.dup,
                                                       :strip,
                                                       @strip_whitespace_keys)
    end

    private

    def filtered_hash_by_meth_for_keys(conditions, method, condition_keys)
      condition_keys.each do |key|
        value = conditions[key]
        conditions[key] = value.send(method) if value.respond_to?(method)
      end

      conditions
    end

    def stringify_params(conditions)
      return conditions unless conditions.is_a? Hash
      conditions.each { |k, v| conditions[k] = v.to_s }
    end
  end
end
