# frozen_string_literal: true

require 'openssl'

module Negroni
  # Generates a secure token.
  class TokenGenerator
    # Creates a new token generator.
    #
    # @param key_generator [ActiveSupport::CachingKeyGenerator] the generator
    # @param digest [String, Symbol] The digest to use.  Must be a stringified
    # version of an OpenSSL::Digest subclass.  Default: `SHA256`.
    def initialize(key_generator = nil, digest = 'SHA256', secret_key: nil)
      @key_generator = if secret_key
                         ActiveSupport::CachingKeyGenerator.new(
                           ActiveSupport::KeyGenerator.new(secret_key)
                         )
                       elsif key_generator
                         key_generator
                       else
                         raise ArgumentError, 'Unexpected arguments!'
                       end
      @digest = digest
    end

    # Digest a key.
    #
    # @param _klass [Class] the class for which to digest the value.
    # @param column [String] the column name to digest.
    # @param value [#to_s] the value to digest.
    #
    # @return [String] the digested value.
    def digest(_klass, column, value)
      return unless value.present?
      OpenSSL::HMAC.hexdigest(@digest, key_for(column), value.to_s)
    end

    # Generate a key.
    #
    # @param klass [Class] the class for which to generate the key.
    # @param column [String] the column name to generate a key for.
    def generate(klass, column)
      key = key_for(column)

      loop do
        raw = Negroni.friendly_token
        enc = OpenSSL::HMAC.hexdigest(@digest, key, raw)
        break [raw, enc] unless klass.to_adapter.find_first(column => enc)
      end
    end

    private

    def key_for(column)
      @key_generator.generate_key "Negroni #{column}"
    end
  end
end
